import synapse.lib.module as s_module

class RiskModule(s_module.CoreModule):

    def getModelDefs(self):

        modl = {
            'types': (
                ('risk:vuln', ('guid', {}), {
                    'doc': 'A unique vulnerability.',
                }),
                ('risk:hasvuln', ('guid', {}), {
                    'doc': 'An instance of a vulnerability present in a target.',
                }),
                ('risk:attack', ('guid', {}), {
                    'doc': 'An instance of an actor attacking a target.'
                }),
            ),
            'forms': (
                ('risk:vuln', {}, (
                    ('name', ('str', {}), {
                        'doc': 'A user specified name for the vulnerability.',
                    }),

                    ('type', ('str', {}), {
                        'doc': 'A user specified type for the vulnerability.',
                    }),
                    ('desc', ('str', {}), {
                        'doc': 'A description of the vulnerability.',
                    }),
                    ('cve', ('it:sec:cve', {}), {
                        'doc': 'The CVE ID of the vulnerability.',
                    }),
                )),

                ('risk:hasvuln', {}, (
                    ('vuln', ('risk:vuln', {}), {
                        'doc': 'The vulnerability present in the target.'
                    }),
                    ('person', ('ps:person', {}), {
                        'doc': 'The vulnerable person.',
                    }),
                    ('org', ('ou:org', {}), {
                        'doc': 'The vulnerable org.',
                    }),
                    ('place', ('geo:place', {}), {
                        'doc': 'The vulnerable place.',
                    }),
                    ('software', ('it:prod:softver', {}), {
                        'doc': 'The vulnerable software.',
                    }),
                    ('spec', ('mat:spec', {}), {
                        'doc': 'The vulnerable material specification.',
                    }),
                    ('item', ('mat:item', {}), {
                        'doc': 'The vulnerable material item.',
                    }),
                )),

                ('risk:attack', {}, (
                    ('time', ('time', {}), {
                        'doc': 'Set if the time of the attack is known.',
                    }),
                    ('success', ('bool', {}), {
                        'doc': 'Set if the attack was known to have succeeded or not.',
                    }),
                    ('targeted', ('bool', {}), {
                        'doc': 'Set if the attack was assessed to be targeted or not.',
                    }),
                    ('campaign', ('ou:campaign', {}), {
                        'doc': 'Set if the attack was part of a larger campaign.',
                    }),
                    ('prev', ('risk:attack', {}), {
                        'doc': 'The previous/parent attack in a list or hierarchy.',
                    }),
                    ('actor:org', ('ou:org', {}), {
                        'doc': 'The org that carried out the attack.',
                    }),
                    ('actor:person', ('ps:person', {}), {
                        'doc': 'The person that carried out the attack.',
                    }),
                    ('target:org', ('ou:org', {}), {
                        'doc': 'The org was the target of the attack.',
                    }),
                    ('target:host', ('it:host', {}), {
                        'doc': 'The host was the target of the attack.',
                    }),
                    ('target:person', ('ps:person', {}), {
                        'doc': 'The person was the target of the attack.',
                    }),
                    ('via:ipv4', ('inet:ipv4', {}), {
                        'doc': 'The target host was contacted via the IPv4 address.',
                    }),
                    ('via:ipv6', ('inet:ipv6', {}), {
                        'doc': 'The target host was contacted via the IPv6 address.',
                    }),
                    ('via:email', ('inet:email', {}), {
                        'doc': 'The target person/org was contacted via the email address.',
                    }),
                    ('via:phone', ('tel:phone', {}), {
                        'doc': 'The target person/org was contacted via the phone number.',
                    }),
                    ('used:vuln', ('risk:vuln', {}), {
                        'doc': 'The actor used the vuln in the attack.',
                    }),
                    ('used:url', ('inet:url', {}), {
                        'doc': 'The actor used the url in the attack.',
                    }),
                    ('used:host', ('it:host', {}), {
                        'doc': 'The actor used the host in the attack.',
                    }),
                    ('used:email', ('inet:email', {}), {
                        'doc': 'The actor used the email in the attack.',
                    }),
                    ('used:file', ('file:bytes', {}), {
                        'doc': 'The actor used the file in the attack.',
                    }),
                    ('used:server', ('inet:server', {}), {
                        'doc': 'The actor used the server in the attack.',
                    }),
                    ('used:software', ('it:prod:softver', {}), {
                        'doc': 'The actor used the software in the attack.',
                    }),
                )),
            ),
        }
        name = 'risk'
        return ((name, modl), )



import os
import sys
import time
import shutil
import fnmatch
import logging
import argparse

import lmdb

import synapse.common as s_common

logger = logging.getLogger(__name__)

def backup(srcdir, dstdir, skipdirs=None, compact=True):
    '''
    Create a backup of a Synapse application.

    Args:
        srcdir (str): Path to the directory to backup.
        dstdir (str): Path to backup target directory.
        skipdirs (list or None): Optional list of relative directory name glob patterns to exclude from the backup.
        compact (bool): Whether to optimize storage while copying to the destination.
    '''
    tick = s_common.now()

    srcdir = s_common.reqdir(srcdir)
    dstdir = s_common.gendir(dstdir)

    if skipdirs is None:
        skipdirs = []

    # Always avoid backing up temporary directories
    skipdirs.append('**/tmp')

    logger.info(f'Starting backup of [{srcdir}]')
    logger.info(f'Destination dir: [{dstdir}]')

    for root, dnames, fnames in os.walk(srcdir, topdown=True):

        relpath = os.path.relpath(root, start=srcdir)

        for name in list(dnames):

            srcpath = s_common.genpath(root, name)

            relname = os.path.join(relpath, name)

            if any([fnmatch.fnmatch(relname, pattern) for pattern in skipdirs]):
                logger.info(f'skipping dir:{srcpath}')
                dnames.remove(name)
                continue

            dstpath = s_common.genpath(dstdir, relname)

            if name.endswith('.lmdb'):
                dnames.remove(name)
                backup_lmdb(srcpath, dstpath, compact)
                continue

            logger.info(f'making dir:{dstpath}')
            s_common.gendir(dstpath)

        for name in fnames:

            srcpath = s_common.genpath(root, name)
            # skip unix sockets etc...
            if not os.path.isfile(srcpath):
                continue

            dstpath = s_common.genpath(dstdir, relpath, name)
            logger.info(f'copying: {srcpath} -> {dstpath}')
            shutil.copy(srcpath, dstpath)

    tock = s_common.now()

    logger.info(f'Backup complete. Took [{tock-tick:.2f}] for [{srcdir}]')
    return

def backup_lmdb(envpath, dstdir, compact=True):

    datafile = os.path.join(envpath, 'data.mdb')
    stat = os.stat(datafile)
    map_size = stat.st_size

    env = lmdb.open(
        envpath,
        map_size=map_size,
        subdir=True,
        max_dbs=256,
        create=False,
        readonly='READ'
    )

    tick = time.time()

    s_common.gendir(dstdir)

    env.copy(dstdir, compact=compact)

    tock = time.time()
    logger.info(f'backup took: {tock-tick:.2f} seconds')
    env.close()

def main(argv):
    args = parse_args(argv)
    backup(args.srcdir, args.dstdir, args.skipdirs)
    return 0

def parse_args(argv):
    desc = 'Create an optimized backup of a Synapse directory.'
    parser = argparse.ArgumentParser('synapse.tools.backup', description=desc)
    parser.add_argument('srcdir', help='Path to the Synapse directory to backup.')
    parser.add_argument('dstdir', help='Path to the backup target directory.')
    parser.add_argument('--skipdirs', nargs='+',
                        help='Glob patterns of relative directory names to exclude from the backup.')
    args = parser.parse_args(argv)
    return args

def _main(argv):  # pragma: no cover
    s_common.setlogging(logger, defval='DEBUG')
    return main(argv)

if __name__ == '__main__':  # pragma: no cover
    sys.exit(_main(sys.argv[1:]))



import synapse.lib.module as s_module

class OuModule(s_module.CoreModule):
    def getModelDefs(self):
        modl = {
            'types': (
                ('ou:sic', ('str', {'regex': r'^[0-9]{4}$'}), {
                    'doc': 'The four digit Standard Industrial Classification Code.',
                    'ex': '0111',
                }),
                ('ou:naics', ('str', {'regex': r'^[1-9][0-9]{4}[0-9]?$'}), {
                    'doc': 'The five or six digit North American Industry Classification System code.',
                    'ex': '541715',
                }),
                ('ou:org', ('guid', {}), {
                    'doc': 'A GUID for a human organization such as a company or military unit.',
                }),
                ('ou:alias', ('str', {'lower': True, 'regex': r'^[0-9a-z_]+$'}), {
                    'doc': 'An alias for the org GUID.',
                    'ex': 'vertexproject',
                }),
                ('ou:hasalias', ('comp', {'fields': (('org', 'ou:org'), ('alias', 'ou:alias'))}), {
                    'doc': 'The knowledge that an organization has an alias.',
                }),
                ('ou:orgnet4', ('comp', {'fields': (('org', 'ou:org'), ('net', 'inet:net4'))}), {
                    'doc': "An organization's IPv4 netblock.",
                }),
                ('ou:orgnet6', ('comp', {'fields': (('org', 'ou:org'), ('net', 'inet:net6'))}), {
                    'doc': "An organization's IPv6 netblock.",
                }),
                ('ou:name', ('str', {'lower': True, 'strip': True}), {
                    'doc': 'The name of an organization. This may be a formal name or informal name of the '
                           'organization.',
                    'ex': 'acme corporation',
                }),
                ('ou:member', ('comp', {'fields': (('org', 'ou:org'), ('person', 'ps:person'))}), {
                    'doc': 'A person who is (or was) a member of an organization.',
                }),
                ('ou:suborg', ('comp', {'fields': (('org', 'ou:org'), ('sub', 'ou:org'))}), {
                    'doc': 'Any parent/child relationship between two orgs. May represent ownership, organizational structure, etc.',
                }),
                ('ou:org:has', ('comp', {'fields': (('org', 'ou:org'), ('node', 'ndef'))}), {
                    'doc': 'An org owns, controls, or has exclusive use of an object or resource, '
                           'potentially during a specific period of time.',
                }),
                ('ou:user', ('comp', {'fields': (('org', 'ou:org'), ('user', 'inet:user'))}), {
                    'doc': 'A user name within an organization.',
                }),
                ('ou:meet', ('guid', {}), {
                    'doc': 'An informal meeting of people which has no title or sponsor.  See also: ou:conference.',
                }),
                ('ou:meet:attendee', ('comp', {'fields': (('meet', 'ou:meet'), ('person', 'ps:person'))}), {
                    'doc': 'Represents a person attending a meeting represented by an ou:meet node.',
                }),
                ('ou:conference', ('guid', {}), {
                    'doc': 'A conference with a name and sponsoring org.',
                }),
                ('ou:conference:attendee', ('comp', {'fields': (('conference', 'ou:conference'), ('person', 'ps:person'))}), {
                    'doc': 'Represents a person attending a conference represented by an ou:conference node.',
                }),
                ('ou:conference:event', ('guid', {}), {
                    'doc': 'A conference event with a name and associated conference.',
                }),
                ('ou:conference:event:attendee', ('comp', {'fields': (('conference', 'ou:conference:event'), ('person', 'ps:person'))}), {
                    'doc': 'Represents a person attending a conference event represented by an ou:conference:event node.',
                }),
                ('ou:goal', ('guid', {}), {
                    'doc': 'An assessed or stated goal which may be abstract or org specific.',
                }),
                ('ou:hasgoal', ('comp', {'fields': (('org', 'ou:org'), ('goal', 'ou:goal'))}), {
                    'doc': 'An org has an assessed or stated goal.',
                }),
                ('ou:campaign', ('guid', {}), {
                    'doc': 'Represents an orgs activity in pursuit of a goal.',
                }),
            ),
            'forms': (
                ('ou:goal', {}, (
                    ('name', ('str', {}), {
                        'doc': 'A terse name for the goal.',
                    }),
                    ('type', ('str', {}), {
                        'doc': 'A user specified goal type.',
                    }),
                    ('desc', ('str', {}), {
                        'doc': 'A description of the goal.',
                    }),
                    ('prev', ('ou:goal', {}), {
                        'doc': 'The previous/parent goal in a list or hierarchy.',
                    }),
                )),
                ('ou:hasgoal', {}, (
                    ('org', ('ou:org', {}), {
                        'doc': 'The org which has the goal.',
                    }),
                    ('goal', ('ou:goal', {}), {
                        'doc': 'The goal which the org has.',
                    }),
                    ('stated', ('bool', {}), {
                        'doc': 'Set to true/false if the goal is known to be self stated.',
                    }),
                    ('window', ('ival', {}), {
                        'doc': 'Set if a goal has a limited time window.',
                    }),
                )),
                ('ou:campaign', {}, (
                    ('org', ('ou:org', {}), {
                        'doc': 'The org carrying out the campaign.',
                    }),
                    ('goal', ('ou:goal', {}), {
                        'doc': 'The assessed primary goal of the campaign.',
                    }),
                    ('goals', ('array', {'type': 'ou:goal'}), {
                        'doc': 'Additional assessed goals of the campaign.',
                    }),
                    ('name', ('str', {}), {
                        'doc': 'A terse name of the campaign.',
                    }),
                    ('type', ('str', {}), {
                        'doc': 'A user specified campaign type.',
                    }),
                    ('desc', ('str', {}), {
                        'doc': 'A description of the campaign.',
                    }),
                )),
                ('ou:org', {}, (
                    ('loc', ('loc', {}), {
                        'doc': 'Location for an organization.'
                    }),
                    ('name', ('ou:name', {}), {
                        'doc': 'The localized name of an organization.',
                    }),
                    ('names', ('array', {'type': 'ou:name'}), {
                       'doc': 'A list of alternate names for the organization.',
                    }),
                    ('alias', ('ou:alias', {}), {
                        'doc': 'The default alias for an organization.'
                    }),
                    ('phone', ('tel:phone', {}), {
                        'doc': 'The primary phone number for the organization.',
                    }),
                    ('sic', ('ou:sic', {}), {
                        'doc': 'The Standard Industrial Classification code for the organization.',
                    }),
                    ('naics', ('ou:naics', {}), {
                        'doc': 'The North American Industry Classification System code for the organization.',
                    }),
                    ('us:cage', ('gov:us:cage', {}), {
                        'doc': 'The Commercial and Government Entity (CAGE) code for the organization.',
                    }),
                    ('founded', ('time', {}), {
                        'doc': 'The date on which the org was founded.'}),
                    ('dissolved', ('time', {}), {
                        'doc': 'The date on which the org was dissolved.'}),
                    ('url', ('inet:url', {}), {
                        'doc': 'The primary url for the organization.',
                    }),
                )),
                ('ou:name', {}, ()),
                ('ou:hasalias', {}, (
                    ('org', ('ou:org', {}), {
                        'ro': True,
                        'doc': 'The org guid which has the alias.',
                    }),
                    ('alias', ('ou:alias', {}), {
                        'ro': True,
                        'doc': 'Alias for the organization.',
                    }),
                )),
                ('ou:orgnet4', {}, (
                    ('org', ('ou:org', {}), {
                        'ro': True,
                        'doc': 'The org guid which owns the netblock.',
                    }),
                    ('net', ('inet:net4', {}), {
                        'ro': True,
                        'doc': 'Netblock owned by the organization.',
                    }),
                    ('name', ('str', {'lower': True, 'strip': True}), {
                        'doc': 'The name that the organization assigns to this netblock.'
                    }),
                )),
                ('ou:orgnet6', {}, (
                    ('org', ('ou:org', {}), {
                        'ro': True,
                        'doc': 'The org guid which owns the netblock.',
                    }),
                    ('net', ('inet:net6', {}), {
                        'ro': True,
                        'doc': 'Netblock owned by the organization.',
                    }),
                    ('name', ('str', {'lower': True, 'strip': True}), {
                        'doc': 'The name that the organization assigns to this netblock.'
                    }),
                )),
                ('ou:member', {}, (
                    ('org', ('ou:org', {}), {
                        'ro': True,
                        'doc': 'The GUID of the org the person is a member of.',
                    }),
                    ('person', ('ps:person', {}), {
                        'ro': True,
                        'doc': 'The GUID of the person that is a member of an org.',
                    }),
                    ('title', ('str', {'lower': True, 'strip': True}), {
                        'doc': 'The persons normalized title.'
                    }),
                    ('start', ('time', {'ismin': True}), {
                        'doc': 'Earliest known association of the person with the org.',
                    }),
                    ('end', ('time', {'ismax': True}), {
                        'doc': 'Most recent known association of the person with the org.',
                    })
                )),
                ('ou:suborg', {}, (
                    ('org', ('ou:org', {}), {
                        'ro': True,
                        'doc': 'The org which owns the sub organization.',
                    }),
                    ('sub', ('ou:org', {}), {
                        'ro': True,
                        'doc': 'The sub org which owned by the org.',
                    }),
                    ('perc', ('int', {'min': 0, 'max': 100}), {
                        'doc': 'The optional percentage of sub which is owned by org.',
                    }),
                    ('current', ('bool', {}), {
                        'doc': 'Bool indicating if the suborg relationship still current.',
                    }),
                )),
                ('ou:org:has', {}, (
                    ('org', ('ou:org', {}), {
                        'ro': True,
                        'doc': 'The org who owns or controls the object or resource.',
                    }),
                    ('node', ('ndef', {}), {
                        'ro': True,
                        'doc': 'The object or resource that is owned or controlled by the org.',
                    }),
                    ('node:form', ('str', {}), {
                        'ro': True,
                        'doc': 'The form of the object or resource that is owned or controlled by the org.',
                    }),
                )),
                ('ou:user', {}, (
                    ('org', ('ou:org', {}), {
                        'ro': True,
                        'doc': 'The org guid which owns the netblock.',
                    }),
                    ('user', ('inet:user', {}), {
                        'ro': True,
                        'doc': 'The username associated with the organization.',
                    }),
                )),
                ('ou:meet', {}, (
                    ('name', ('str', {'lower': True}), {
                        'doc': 'A human friendly name for the meeting.',
                    }),
                    ('start', ('time', {}), {
                        'doc': 'The date / time the meet starts.',
                    }),
                    ('end', ('time', {}), {
                        'doc': 'The date / time the meet ends.',
                    }),
                    ('place', ('geo:place', ()), {
                        'doc': 'The geo:place node where the meet was held.',
                    }),
                )),
                ('ou:meet:attendee', {}, (
                    ('meet', ('ou:meet', {}), {
                        'ro': True,
                        'doc': 'The meeting which was attended.',
                    }),
                    ('person', ('ps:person', {}), {
                        'ro': True,
                        'doc': 'The person who attended the meeting.',
                    }),
                    ('arrived', ('time', {}), {
                        'doc': 'The time when a person arrived to the meeting.',
                    }),
                    ('departed', ('time', {}), {
                        'doc': 'The time when a person departed from the meeting.',
                    }),
                )),
                ('ou:conference', {}, (
                    ('org', ('ou:org', {}), {
                        'doc': 'The org which created/managed the conference.',
                    }),
                    ('name', ('str', {'lower': True}), {
                        'doc': 'The full name of the conference.',
                        'ex': 'decfon 2017',
                    }),
                    ('desc', ('str', {'lower': True}), {
                        'doc': 'A description of the conference.',
                        'ex': 'annual cybersecurity conference',
                    }),
                    ('base', ('str', {'lower': True, 'strip': True}), {
                        'doc': 'The base name which is shared by all conference instances.',
                        'ex': 'defcon',
                    }),
                    ('start', ('time', {}), {
                        'doc': 'The conference start date / time.',
                    }),
                    ('end', ('time', {}), {
                        'doc': 'The conference end date / time.',
                    }),
                    ('place', ('geo:place', ()), {
                        'doc': 'The geo:place node where the conference was held.',
                    }),
                    ('url', ('inet:url', ()), {
                        'doc': 'The inet:url node for the conference website.',
                    }),
                )),
                ('ou:conference:attendee', {}, (
                    ('conference', ('ou:conference', {}), {
                        'ro': True,
                        'doc': 'The conference which was attended.',
                    }),
                    ('person', ('ps:person', {}), {
                        'ro': True,
                        'doc': 'The person who attended the conference.',
                    }),
                    ('arrived', ('time', {}), {
                        'doc': 'The time when a person arrived to the conference.',
                    }),
                    ('departed', ('time', {}), {
                        'doc': 'The time when a person departed from the conference.',
                    }),
                    ('role:staff', ('bool', {}), {
                        'doc': 'The person worked as staff at the conference.',
                    }),
                    ('role:speaker', ('bool', {}), {
                        'doc': 'The person was a speaker or presenter at the conference.',
                    }),
                    ('roles', ('array', {'type': 'str', 'lower': True}), {
                        'doc': 'List of the roles the person had at the conference.',
                    }),
                 )),
                ('ou:conference:event', {}, (
                    ('conference', ('ou:conference', {}), {
                        'ro': True,
                        'doc': 'The conference to which the event is associated.',
                    }),
                    ('place', ('geo:place', {}), {
                        'doc': 'The geo:place where the event occurred.',
                    }),
                    ('name', ('str', {'lower': True}), {
                        'doc': 'The name of the conference event.',
                        'ex': 'foobar conference dinner',
                    }),
                    ('desc', ('str', {'lower': True}), {
                        'doc': 'A description of the conference event.',
                        'ex': 'foobar conference networking dinner at ridge hotel',
                    }),
                    ('url', ('inet:url', ()), {
                        'doc': 'The inet:url node for the conference event website.',
                    }),
                    ('contact', ('ps:contact', ()), {
                        'doc': 'Contact info for the event.',
                    }),
                    ('start', ('time', {}), {
                        'doc': 'The event start date / time.',
                    }),
                    ('end', ('time', {}), {
                        'doc': 'The event end date / time.',
                    }),
                )),
                ('ou:conference:event:attendee', {}, (

                    ('event', ('ou:conference:event', {}), {
                        'ro': True,
                        'doc': 'The conference event which was attended.',
                    }),
                    ('person', ('ps:person', {}), {
                        'ro': True,
                        'doc': 'The person who attended the conference event.',
                    }),
                    ('arrived', ('time', {}), {
                        'doc': 'The time when a person arrived to the conference event.',
                    }),
                    ('departed', ('time', {}), {
                        'doc': 'The time when a person departed from the conference event.',
                    }),
                    ('roles', ('array', {'type': 'str', 'lower': True}), {
                        'doc': 'List of the roles the person had at the conference event.',
                    }),
                )),
            )
        }

        name = 'ou'
        return ((name, modl),)



import os
import signal
import asyncio
import logging
import multiprocessing

import synapse.exc as s_exc
import synapse.glob as s_glob
import synapse.cortex as s_cortex

import synapse.lib.coro as s_coro
import synapse.lib.link as s_link
import synapse.lib.spawn as s_spawn
import synapse.lib.msgpack as s_msgpack
import synapse.lib.lmdbslab as s_lmdbslab

import synapse.tests.utils as s_test
import synapse.tests.test_lib_stormsvc as s_test_svc

logger = logging.getLogger(__name__)

def make_core(dirn, conf, queries, queue, event):
    '''
    Multiprocessing target for making a Cortex for local use of a SpawnCore instance.
    '''

    async def workloop():
        s_glob.iAmLoop()
        async with await s_cortex.Cortex.anit(dirn=dirn, conf=conf) as core:
            await core.addTagProp('added', ('time', {}), {})
            for q in queries:
                await core.nodes(q)
            core.view.layers[0].layrslab.forcecommit()

            spawninfo = await core.getSpawnInfo()
            queue.put(spawninfo)
            # Don't block the ioloop..
            await s_coro.executor(event.wait)

    asyncio.run(workloop())

class CoreSpawnTest(s_test.SynTest):

    async def test_spawncore(self):
        # This test makes a real Cortex in a remote process, and then
        # gets the spawninfo from that real Cortex in order to make a
        # local SpawnCore. This avoids the problem of being unable to
        # open lmdb environments multiple times by the same process
        # and allows direct testing of the SpawnCore object.

        mpctx = multiprocessing.get_context('spawn')
        queue = mpctx.Queue()
        event = mpctx.Event()

        conf = {
            'storm:log': True,
            'storm:log:level': logging.INFO,
            'modules': [('synapse.tests.utils.TestModule', {})],
        }
        queries = [
            '[test:str="Cortex from the aether!"]',
        ]
        with self.getTestDir() as dirn:
            args = (dirn, conf, queries, queue, event)
            proc = mpctx.Process(target=make_core, args=args)
            try:
                proc.start()
                spawninfo = queue.get(timeout=30)

                async with await s_spawn.SpawnCore.anit(spawninfo) as core:
                    root = await core.auth.getUserByName('root')
                    q = '''test:str
                    $lib.print($lib.str.format("{n}", n=$node.repr()))
                    | limit 1'''
                    item = {
                        'user': root.iden,
                        'view': list(core.views.keys())[0],
                        'storm': {
                            'query': q,
                            'opts': None,
                        }
                    }

                    # Test the storm implementation used by spawncore
                    msgs = await s_test.alist(s_spawn.storm(core, item))
                    podes = [m[1] for m in msgs if m[0] == 'node']
                    e = 'Cortex from the aether!'
                    self.len(1, podes)
                    self.eq(podes[0][0], ('test:str', e))
                    self.stormIsInPrint(e, msgs)

                    # Direct test of the _innerloop code.
                    todo = mpctx.Queue()
                    done = mpctx.Queue()

                    # Test poison - this would cause the corework to exit
                    todo.put(None)
                    self.none(await s_spawn._innerloop(core, todo, done))

                    # Test a real item with a link associated with it. This ends
                    # up getting a bunch of telepath message directly.
                    todo_item = item.copy()
                    link0, sock0 = await s_link.linksock()
                    todo_item['link'] = link0.getSpawnInfo()
                    todo.put(todo_item)
                    self.true(await s_spawn._innerloop(core, todo, done))
                    resp = done.get(timeout=12)
                    self.false(resp)
                    buf0 = sock0.recv(1024 * 16)
                    unpk = s_msgpack.Unpk()
                    msgs = [msg for (offset, msg) in unpk.feed(buf0)]
                    self.eq({'t2:genr', 't2:yield'},
                            {m[0] for m in msgs})

                    await link0.fini()  # We're done with the link now
                    todo.close()
                    done.close()

                # Test the workloop directly - this again just gets telepath
                # messages back. This does use poison to kill the workloop.
                todo = mpctx.Queue()
                done = mpctx.Queue()

                task = asyncio.create_task(s_spawn._workloop(spawninfo, todo, done))
                await asyncio.sleep(0.01)
                link1, sock1 = await s_link.linksock()
                todo_item = item.copy()
                todo_item['link'] = link1.getSpawnInfo()
                todo.put(todo_item)
                # Don't block the IO loop!
                resp = await s_coro.executor(done.get, timeout=12)
                self.false(resp)
                buf0 = sock1.recv(1024 * 16)
                unpk = s_msgpack.Unpk()
                msgs = [msg for (offset, msg) in unpk.feed(buf0)]
                self.eq({'t2:genr', 't2:yield'},
                        {m[0] for m in msgs})
                await link1.fini()  # We're done with the link now
                # Poison the queue - this should close the task
                todo.put(None)
                self.none(await asyncio.wait_for(task, timeout=12))

                todo.close()
                done.close()

            finally:

                queue.close()
                event.set()
                proc.join(12)

    async def test_cortex_spawn_telepath(self):
        conf = {
            'storm:log': True,
            'storm:log:level': logging.INFO,
        }

        async with self.getTestCore(conf=conf) as core:
            pkgdef = {
                'name': 'spawn',
                'version': (0, 0, 1),
                'commands': (
                    {
                        'name': 'passthrough',
                        'desc': 'passthrough input nodes and print their ndef',
                        'storm': '$lib.print($node.ndef())',
                    },
                ),
            }

            await self.runCoreNodes(core, '[ inet:dns:a=(vertex.link, 1.2.3.4) ] -> inet:ipv4 [ :asn=0 ]')

            async with core.getLocalProxy() as prox:

                opts = {'spawn': True}

                # check that regular node lifting / pivoting works
                msgs = await prox.storm('inet:fqdn=vertex.link -> inet:dns:a -> inet:ipv4', opts=opts).list()
                podes = [m[1] for m in msgs if m[0] == 'node']
                self.len(1, podes)
                self.eq(podes[0][0], ('inet:ipv4', 0x01020304))

                # test that runt node lifting works
                msgs = await prox.storm('syn:prop=inet:dns:a:fqdn :form -> syn:form', opts=opts).list()
                podes = [m[1] for m in msgs if m[0] == 'node']
                self.len(1, podes)
                self.eq(podes[0][0], ('syn:form', 'inet:dns:a'))

                # make sure node creation fails cleanly
                msgs = await prox.storm('[ inet:email=visi@vertex.link ]', opts=opts).list()
                errs = [m[1] for m in msgs if m[0] == 'err']
                self.eq(errs[0][0], 'IsReadOnly')

                # make sure storm commands are loaded
                msgs = await prox.storm('inet:ipv4=1.2.3.4 | limit 1', opts=opts).list()
                podes = [m[1] for m in msgs if m[0] == 'node']
                self.len(1, podes)
                self.eq(podes[0][0], ('inet:ipv4', 0x01020304))

                # make sure graph rules work
                msgs = await prox.storm('inet:dns:a', opts={'spawn': True, 'graph': True}).list()
                podes = [m[1] for m in msgs if m[0] == 'node']
                ndefs = list(sorted(p[0] for p in podes))

                self.eq(ndefs, (
                    ('inet:asn', 0),
                    ('inet:dns:a', ('vertex.link', 16909060)),
                    ('inet:fqdn', 'link'),
                    ('inet:fqdn', 'vertex.link'),
                    ('inet:ipv4', 16909060),
                ))

                # Test a python cmd that came in via a ctor
                msgs = await prox.storm('inet:ipv4=1.2.3.4 | testcmd', opts=opts).list()
                self.stormIsInPrint("testcmd: ('inet:ipv4', 16909060)", msgs)
                podes = [m[1] for m in msgs if m[0] == 'node']
                self.len(1, podes)

                # Test a macro execution
                msgs = await prox.storm('macro.set macrotest ${ inet:ipv4 }', opts=opts).list()
                self.stormIsInPrint('Set macro: macrotest', msgs)
                msgs = await prox.storm('macro.list', opts=opts).list()
                self.stormIsInPrint('macrotest', msgs)
                self.stormIsInPrint('owner: root', msgs)
                msgs = await prox.storm('macro.exec macrotest', opts=opts).list()
                podes = [m[1] for m in msgs if m[0] == 'node']
                self.len(1, podes)
                self.eq(podes[0][0], ('inet:ipv4', 0x01020304))
                msgs = await prox.storm('macro.del macrotest', opts=opts).list()
                self.stormIsInPrint('Removed macro: macrotest', msgs)
                msgs = await prox.storm('macro.exec macrotest', opts=opts).list()
                self.stormIsInErr('Macro name not found: macrotest', msgs)

                # Test a simple stormlib command
                msgs = await prox.storm('$lib.print("hello")', opts=opts).list()
                self.stormIsInPrint("hello", msgs)

                # test a complex stormlib command using lib deferences
                marsopts = {'spawn': True, 'vars': {'world': 'mars'}}
                q = '$lib.print($lib.str.format("hello {world}", world=$world))'
                msgs = await prox.storm(q, opts=marsopts).list()
                self.stormIsInPrint("hello mars", msgs)

                # Model deference off of the snap via stormtypes
                q = '''$valu=$lib.time.format('200103040516', '%Y %m %d')
                $lib.print($valu)
                '''
                msgs = await prox.storm(q, opts=opts).list()
                self.stormIsInPrint('2001 03 04', msgs)

                # Test sleeps / fires from a spawnproc
                q = '''$tick=$lib.time.now()
                $lib.time.sleep(0.1)
                $tock=$lib.time.now()
                $lib.fire(took, tick=$tick, tock=$tock)
                '''
                msgs = await prox.storm(q, opts=opts).list()
                fires = [m[1] for m in msgs if m[0] == 'storm:fire']
                self.len(1, fires)
                fire_data = fires[0].get('data')
                self.ne(fire_data.get('tick'), fire_data.get('tock'))

                # Add a stormpkg - this should fini the spawnpool spawnprocs
                procs = [p for p in core.spawnpool.spawns.values()]
                self.isin(len(procs), (1, 2, 3))

                await core.addStormPkg(pkgdef)

                for proc in procs:
                    self.true(await proc.waitfini(6))

                self.len(0, core.spawnpool.spawnq)
                self.len(0, core.spawnpool.spawns)

                # Test a pure storm commands
                msgs = await prox.storm('inet:fqdn=vertex.link | passthrough', opts=opts).list()
                self.stormIsInPrint("('inet:fqdn', 'vertex.link')", msgs)

                # No guarantee that we've gotten the proc back into
                # the pool so we cannot check the size of spawnq
                self.len(1, core.spawnpool.spawns)

                # Test launching a bunch of spawn queries at the same time
                donecount = 0

                await self.runCoreNodes(core, '[test:int=1]')
                # force a commit
                await s_lmdbslab.Slab.syncLoopOnce()

                async def taskfunc(i):
                    nonlocal donecount
                    msgs = await prox.storm('test:int=1 | sleep 3', opts=opts).list()
                    if len(msgs) == 3:
                        donecount += 1

                n = 4
                tasks = [taskfunc(i) for i in range(n)]
                try:
                    await asyncio.wait_for(asyncio.gather(*tasks), timeout=80)
                except asyncio.TimeoutError:
                    self.fail('Timeout awaiting for spawn tasks to finish.')

                self.eq(donecount, n)

                # test a remote boss kill of the client side task
                logger.info('telepath ps/kill test.')
                evnt = asyncio.Event()
                msgs = {'msgs': []}

                tf2opts = {'spawn': True, 'vars': {'hehe': 'haha'}}

                async def taskfunc2():
                    async for mesg in prox.storm('test:int=1 | sleep 15', opts=tf2opts):
                        msgs['msgs'].append(mesg)
                        if mesg[0] == 'node':
                            evnt.set()
                    return True

                victimproc = core.spawnpool.spawnq[0]  # type: s_spawn.SpawnProc
                fut = core.schedCoro(taskfunc2())
                self.true(await asyncio.wait_for(evnt.wait(), timeout=6))
                tasks = await prox.ps()
                new_idens = [task.get('iden') for task in tasks]
                self.len(1, new_idens)
                await prox.kill(new_idens[0])

                # Ensure that opts were passed into the task data without spawn: True set
                task = [task for task in tasks if task.get('iden') == new_idens[0]][0]
                self.none(task.get('info').get('opts').get('spawn'))

                # Ensure the task cancellation tore down the spawnproc
                self.true(await victimproc.waitfini(6))

                resp = await fut
                self.true(resp)
                # We did not get a fini messages since the proc was killed
                self.eq({m[0] for m in msgs.get('msgs')}, {'init', 'node'})

                # test kill -9 ing a spawn proc
                logger.info('sigkill test')
                assert len(core.spawnpool.spawnq)
                victimproc = core.spawnpool.spawnq[0]  # type: s_spawn.SpawnProc
                victimpid = victimproc.proc.pid
                sig = signal.SIGKILL

                async def taskfunc3():
                    retn = await prox.storm('test:int=1 | sleep 15', opts=opts).list()
                    return retn

                fut = core.schedCoro(taskfunc3())
                await asyncio.sleep(1)
                os.kill(victimpid, sig)
                self.true(await victimproc.waitfini(6))
                msgs = await fut
                # We did not get a fini messages since the proc was killed
                self.eq({m[0] for m in msgs}, {'init', 'node'})

    async def test_queues(self):
        conf = {
            'storm:log': True,
            'storm:log:level': logging.INFO,
        }

        # Largely mimics test_storm_lib_queue
        async with self.getTestCore(conf=conf) as core:
            opts = {'spawn': True}

            async with core.getLocalProxy() as prox:

                msgs = await prox.storm('queue.add visi', opts=opts).list()
                self.stormIsInPrint('queue added: visi', msgs)

                with self.raises(s_exc.DupName):
                    await core.nodes('queue.add visi')

                msgs = await prox.storm('queue.list', opts=opts).list()
                self.stormIsInPrint('Storm queue list:', msgs)
                self.stormIsInPrint('visi', msgs)

                # Make a node and put it into the queue
                q = '$q = $lib.queue.get(visi) [ inet:ipv4=1.2.3.4 ] $q.put( $node.repr() )'
                nodes = await core.nodes(q)
                self.len(1, nodes)

                await s_lmdbslab.Slab.syncLoopOnce()

                q = '$q = $lib.queue.get(visi) ($offs, $ipv4) = $q.get(0) inet:ipv4=$ipv4'
                msgs = await prox.storm(q, opts=opts).list()

                podes = [m[1] for m in msgs if m[0] == 'node']
                self.len(1, podes)
                self.eq(podes[0][0], ('inet:ipv4', 0x01020304))

                # test iter use case
                q = '$q = $lib.queue.add(blah) [ inet:ipv4=1.2.3.4 inet:ipv4=5.5.5.5 ] $q.put( $node.repr() )'
                nodes = await core.nodes(q)
                self.len(2, nodes)

                await s_lmdbslab.Slab.syncLoopOnce()

                # Put a value into the queue that doesn't exist in the cortex so the lift can nop
                q = '$q = $lib.queue.get(blah) $q.put("8.8.8.8")'
                msgs = await prox.storm(q, opts=opts).list()

                msgs = await prox.storm('''
                    $q = $lib.queue.get(blah)
                    for ($offs, $ipv4) in $q.gets(0, cull=$lib.false, wait=$lib.false) {
                        inet:ipv4=$ipv4
                    }
                ''', opts=opts).list()
                podes = [m[1] for m in msgs if m[0] == 'node']
                self.len(2, podes)

                msgs = await prox.storm('''
                    $q = $lib.queue.get(blah)
                    for ($offs, $ipv4) in $q.gets(wait=$lib.false) {
                        inet:ipv4=$ipv4
                        $q.cull($offs)
                    }
                ''', opts=opts).list()
                podes = [m[1] for m in msgs if m[0] == 'node']
                self.len(2, podes)

                q = '''$q = $lib.queue.get(blah)
                for ($offs, $ipv4) in $q.gets(wait=0) {
                    inet:ipv4=$ipv4
                }'''
                msgs = await prox.storm(q, opts=opts).list()
                podes = [m[1] for m in msgs if m[0] == 'node']
                self.len(0, podes)

                msgs = await prox.storm('queue.del visi', opts=opts).list()
                self.stormIsInPrint('queue removed: visi', msgs)

                with self.raises(s_exc.NoSuchName):
                    await core.nodes('queue.del visi')

                msgs = await prox.storm('$lib.queue.get(newp).get()', opts=opts).list()
                # err = msgs[-2]
                errs = [m[1] for m in msgs if m[0] == 'err']
                self.len(1, errs)
                self.eq(errs[0][0], 'NoSuchName')

                # Attempting to use a queue to make nodes in spawn town fails.
                await core.nodes('''
                    $doit = $lib.queue.add(doit)
                    $doit.puts((foo,bar))
                ''')
                q = 'for ($offs, $name) in $lib.queue.get(doit).gets(size=2) { [test:str=$name] }'
                msgs = await prox.storm(q, opts=opts).list()
                errs = [m[1] for m in msgs if m[0] == 'err']
                self.len(1, errs)
                self.eq(errs[0][0], 'IsReadOnly')

            # test other users who have access to this queue can do things to it
            async with core.getLocalProxy() as root:
                # add users
                await root.addUser('synapse')
                await root.addUser('wootuser')

                synu = await core.auth.getUserByName('synapse')
                woot = await core.auth.getUserByName('wootuser')

                async with core.getLocalProxy(user='synapse') as prox:
                    msgs = await prox.storm('queue.add synq', opts=opts).list()
                    errs = [m[1] for m in msgs if m[0] == 'err']
                    self.len(1, errs)
                    self.eq(errs[0][0], 'AuthDeny')

                    rule = (True, ('queue', 'add'))
                    await synu.addRule(rule)
                    msgs = await prox.storm('queue.add synq', opts=opts).list()
                    self.stormIsInPrint('queue added: synq', msgs)

                    rule = (True, ('queue', 'put'))
                    await synu.addRule(rule, gateiden='queue:synq')

                    q = '$q = $lib.queue.get(synq) $q.puts((bar, baz))'
                    msgs = await prox.storm(q, opts=opts).list()

                    # Ensure that the data was put into the queue by the spawnproc
                    q = '$q = $lib.queue.get(synq) $lib.print($q.get(wait=$lib.false, cull=$lib.false))'
                    msgs = await core.stormlist(q)
                    self.stormIsInPrint("(0, 'bar')", msgs)

                async with core.getLocalProxy(user='wootuser') as prox:
                    # now let's see our other user fail to add things
                    msgs = await prox.storm('$lib.queue.get(synq).get()', opts=opts).list()
                    errs = [m[1] for m in msgs if m[0] == 'err']
                    self.len(1, errs)
                    self.eq(errs[0][0], 'AuthDeny')

                    rule = (True, ('queue', 'get'))
                    await woot.addRule(rule, gateiden='queue:synq')

                    q = '$lib.print($lib.queue.get(synq).get(wait=$lib.false))'
                    msgs = await prox.storm(q, opts=opts).list()
                    self.stormIsInPrint("(0, 'bar')", msgs)

                    msgs = await prox.storm('$lib.queue.del(synq)', opts=opts).list()
                    errs = [m[1] for m in msgs if m[0] == 'err']
                    self.len(1, errs)
                    self.eq(errs[0][0], 'AuthDeny')

                    rule = (True, ('queue', 'del'))
                    await woot.addRule(rule, gateiden='queue:synq')

                    msgs = await prox.storm('$lib.queue.del(synq)', opts=opts).list()
                    with self.raises(s_exc.NoSuchName):
                        await core.nodes('$lib.queue.get(synq)')

    async def test_stormpkg(self):
        otherpkg = {
            'name': 'foosball',
            'version': (0, 0, 1),
        }

        stormpkg = {
            'name': 'stormpkg',
            'version': (1, 2, 3)
        }
        conf = {
            'storm:log': True,
            'storm:log:level': logging.INFO,
        }
        async with self.getTestDmon() as dmon:
            dmon.share('real', s_test_svc.RealService())
            host, port = dmon.addr

            lurl = f'tcp://127.0.0.1:{port}/real'
            async with self.getTestCore(conf=conf) as core:

                await core.nodes(f'service.add real {lurl}')
                await core.nodes('$lib.service.wait(real)')
                msgs = await core.stormlist('help')
                self.stormIsInPrint('foobar', msgs)

                async with core.getLocalProxy() as prox:
                    opts = {'spawn': True}

                    msgs = await prox.storm('help', opts=opts).list()
                    self.stormIsInPrint('foobar', msgs)

                    msgs = await prox.storm('pkg.del asdf', opts=opts).list()
                    self.stormIsInPrint('No package names match "asdf". Aborting.', msgs)

                    await core.addStormPkg(otherpkg)
                    msgs = await prox.storm('pkg.list', opts=opts).list()
                    self.stormIsInPrint('foosball', msgs)

                    msgs = await prox.storm(f'pkg.del foosball', opts=opts).list()
                    self.stormIsInPrint('Removing package: foosball', msgs)

                    # Direct add via stormtypes
                    msgs = await prox.storm('$lib.pkg.add($pkg)',
                                            opts={'vars': {'pkg': stormpkg}, 'spawn': True}).list()
                    msgs = await prox.storm('pkg.list', opts=opts).list()
                    self.stormIsInPrint('stormpkg', msgs)

    async def test_spawn_node_data(self):

        # Largely mimics test_storm_node_data
        async with self.getTestCore() as core:
            opts = {'spawn': True}

            async with core.getLocalProxy() as prox:

                await core.nodes('[test:int=10]')
                msgs = await prox.storm('test:int=10 $node.data.set(foo, hehe)', opts=opts).list()
                errs = [m[1] for m in msgs if m[0] == 'err']
                self.eq(errs[0][0], 'IsReadOnly')

                await core.nodes('test:int=10 $node.data.set(foo, hehe)')

                msgs = await prox.storm('test:int $foo=$node.data.get(foo) $lib.print($foo)', opts=opts).list()
                self.stormIsInPrint('hehe', msgs)

                q = 'test:int for $item in $node.data.list() { $lib.print($item) }'
                msgs = await prox.storm(q, opts=opts).list()
                self.stormIsInPrint("('foo', 'hehe')", msgs)

                await core.nodes('test:int=10 $node.data.set(woot, woot)')
                q = 'test:int=10 $node.data.pop(woot)'

                msgs = await prox.storm(q, opts=opts).list()
                errs = [m[1] for m in msgs if m[0] == 'err']
                self.eq(errs[0][0], 'IsReadOnly')

    async def test_model_extensions(self):
        async with self.getTestCoreAndProxy() as (core, prox):
            await core.nodes('[ inet:dns:a=(vertex.link, 1.2.3.4) ]')
            opts = {'spawn': True}
            # Adding model extensions must work
            await core.addFormProp('inet:ipv4', '_woot', ('int', {}), {})
            await core.nodes('[inet:ipv4=1.2.3.4 :_woot=10]')
            await s_lmdbslab.Slab.syncLoopOnce()
            msgs = await prox.storm('inet:ipv4=1.2.3.4', opts=opts).list()
            self.len(3, msgs)
            self.eq(msgs[1][1][1]['props'].get('_woot'), 10)

            msgs = await prox.storm('inet:ipv4:_woot=10', opts=opts).list()
            self.len(3, msgs)
            self.eq(msgs[1][1][1]['props'].get('_woot'), 10)

            # tag props must work
            await prox.addTagProp('added', ('time', {}), {})
            await prox.storm('inet:ipv4=1.2.3.4 [ +#foo.bar:added="2049" ]').list()
            msgs = await prox.storm('inet:ipv4#foo.bar:added', opts=opts).list()
            self.len(3, msgs)

    async def test_spawn_dmon_cmds(self):
        '''
        Copied from test-cortex_storm_lib_dmon_cmds
        '''
        async with self.getTestCoreAndProxy() as (_, prox):
            opts = {'spawn': True}
            await prox.storm('''
                $q = $lib.queue.add(visi)
                $lib.queue.add(boom)

                $lib.dmon.add(${
                    $lib.print('Starting wootdmon')
                    $lib.queue.get(visi).put(blah)
                    for ($offs, $item) in $lib.queue.get(boom).gets(wait=1) {
                        [ inet:ipv4=$item ]
                    }
                }, name=wootdmon)

                for ($offs, $item) in $q.gets(size=1) { $q.cull($offs) }
            ''', opts=opts).list()

            await asyncio.sleep(0)

            # dmon is now fully running
            msgs = await prox.storm('dmon.list', opts=opts).list()
            self.stormIsInPrint('(wootdmon            ): running', msgs)

            msgs = await prox.storm('$lib.dmon.del($ddef.iden)').list()

    async def test_spawn_forked_view(self):
        async with self.getTestCoreAndProxy() as (core, prox):
            await core.nodes('[ test:str=1234 ]')
            mainview = await core.callStorm('$iden=$lib.view.get().pack().iden '
                                            'return ( $iden )')
            forkview = await core.callStorm(f'$fork=$lib.view.get({mainview}).fork() '
                                            f'return ( $fork.pack().iden )')
            await core.nodes('[ test:str=beep ]', {'view': forkview})

            opts = {'spawn': True, 'view': forkview}
            msgs = await prox.storm('test:str $lib.print($node.value()) | spin', opts).list()

            self.stormIsInPrint('1234', msgs)
            self.stormIsInPrint('beep', msgs)
